"use strict";
/*
 * Copyright © 2018 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const sdm_1 = require("@atomist/sdm");
const sdm_core_1 = require("@atomist/sdm-core");
const _ = require("lodash");
const kubeUndeploy_1 = require("./commands/kubeUndeploy");
const kubeDeploy_1 = require("./events/kubeDeploy");
const api_1 = require("./support/api");
/**
 * Register Kubernetes deployment support for provided goals.
 * @param {KubernetesOptions} options
 * @returns {ExtensionPack}
 */
function kubernetesSupport(options = {}) {
    return Object.assign({}, sdm_1.metadata(), { configure: sdm => {
            sdm.addEvent(kubeDeploy_1.kubernetesDeploy);
            if (options.addCommands) {
                sdm.addCommand(kubeUndeploy_1.kubeUndeploy);
            }
            if (sdm_core_1.isInLocalMode()) {
                // set up the kube context
                const context = configureContext(sdm, options);
                // if we are using minikube, set the docker-env too
                if (context === "minikube") {
                    sdm.addStartupListener(DockerEnvStartupListener);
                }
            }
        } });
}
exports.kubernetesSupport = kubernetesSupport;
function configureContext(sdm, options) {
    const kubeConfig = api_1.loadKubeConfig();
    const contexts = kubeConfig.contexts.map((c) => c.context.cluster);
    let context;
    // Assign context
    if (options.context) {
        context = options.context;
    }
    else if (!_.get(sdm, "configuration.sdm.k8.context")) {
        if (contexts.includes("minikube")) {
            context = "minikube";
        }
        else {
            context = kubeConfig["current-context"];
        }
    }
    try {
        // Validate context
        api_1.getKubeConfig(context);
    }
    catch (err) {
        throw new Error(`Failed to load Kubernetes cluster context '${context}'. Available contexts are: ` +
            contexts.join(", "));
    }
    _.set(sdm, "configuration.sdm.k8.context", context);
    return context;
}
const DockerEnvStartupListener = () => __awaiter(this, void 0, void 0, function* () {
    const log = new sdm_1.StringCapturingProgressLog();
    const result = yield automation_client_1.spawnAndWatch({
        command: "minikube",
        args: ["docker-env"],
    }, {}, log, {
        errorFinder: automation_client_1.SuccessIsReturn0ErrorFinder,
        logCommand: false,
    });
    if (result.code === 0) {
        const envVars = log.log.trim().split("\n").filter(l => !l.startsWith("#"));
        envVars.forEach(v => {
            const parts = v.split("=");
            process.env[parts[0].replace(/export /g, "")] = parts[1].replace(/"/g, "");
        });
        automation_client_1.logger.info("Configured local minikube docker env");
    }
});
//# sourceMappingURL=k8.js.map