"use strict";
/*
 * Copyright © 2018 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const sdm_1 = require("@atomist/sdm");
const sdm_core_1 = require("@atomist/sdm-core");
const _ = require("lodash");
const kubernetes_1 = require("../typings/kubernetes");
const deploy_1 = require("./deploy");
const goal_1 = require("./goal");
/**
 * Goal that deploys an application to a Kubernetes cluster by using a direct deploy in SDM local mode or
 * triggers deploy via k8-automation for non-local mode use.
 */
class KubernetesDeploy extends sdm_1.FulfillableGoalWithRegistrations {
    constructor(details, ...dependsOn) {
        super(Object.assign({}, sdm_1.getGoalDefinitionFrom(details, sdm_1.DefaultGoalNameGenerator.generateName("k8-deploy")), { displayName: `deploy${getEnvironmentLabel(details)}`, environment: getEnvironment(details), completedDescription: `Deployed${getEnvironmentLabel(details)}`, failedDescription: `Deployment${getEnvironmentLabel(details)} failed`, waitingForApprovalDescription: `Successfully deployed${getEnvironmentLabel(details)}` }), ...dependsOn);
        this.details = details;
    }
    /**
     * Register a deployment with all required callbacks
     */
    with(registration) {
        if (sdm_core_1.isInLocalMode()) {
            this.addFulfillment({
                name: registration.name,
                goalExecutor: deploy_1.executeKubernetesDeploy(),
                pushTest: registration.pushTest,
            });
        }
        else {
            this.addFulfillment({
                name: "@atomist/k8-automation",
                pushTest: registration.pushTest,
            });
        }
        this.addFulfillmentCallback({
            goal: this,
            callback: kubernetesDataCallback(this, registration),
        });
        return this;
    }
    /**
     * Convenience method to register a deployment
     */
    withDeployment(deploymentData, serviceSpecCreator, deploymentSpecCreator) {
        this.with({
            name: sdm_1.DefaultGoalNameGenerator.generateName("k8-deployer"),
            deploymentData,
            serviceSpecCreator,
            deploymentSpecCreator,
        });
        return this;
    }
    /**
     * Called by the SDM on initialization
     */
    register(sdm) {
        super.register(sdm);
        // Register a startup listener to add the default deployment if no dedicated got registered
        sdm.addStartupListener(() => __awaiter(this, void 0, void 0, function* () {
            if (this.fulfillments.length === 0 &&
                this.callbacks.length === 0) {
                // Register the default deployment
                this.with({ name: "k8-deploy-default", pushTest: sdm_1.AnyPush });
            }
        }));
    }
}
exports.KubernetesDeploy = KubernetesDeploy;
function kubernetesDataCallback(k8Deploy, registration) {
    return (goal, ctx) => __awaiter(this, void 0, void 0, function* () {
        return k8Deploy.sdm.configuration.sdm.projectLoader.doWithProject({
            credentials: ctx.credentials, id: ctx.id, context: ctx.context, readOnly: true,
        }, (p) => __awaiter(this, void 0, void 0, function* () {
            let deploymentSpec = JSON.parse((yield goal_1.readKubernetesSpec(p, "deployment.json")) || "{}");
            if (registration.deploymentSpecCreator) {
                deploymentSpec = yield registration.deploymentSpecCreator(deploymentSpec, goal, ctx);
            }
            let serviceSpec = JSON.parse((yield goal_1.readKubernetesSpec(p, "service.json")) || "{}");
            if (registration.serviceSpecCreator) {
                serviceSpec = yield registration.serviceSpecCreator(serviceSpec, goal, ctx);
            }
            let deploymentData;
            if (registration.deploymentData) {
                deploymentData = yield registration.deploymentData(goal, ctx);
            }
            else {
                deploymentData = yield defaultDeploymentData(p, goal, ctx, k8Deploy);
            }
            if (!deploymentData.environment) {
                deploymentData.environment = k8Deploy.sdm.configuration.environment;
            }
            return Object.assign({}, goal, { data: JSON.stringify(Object.assign({}, JSON.parse(goal.data || "{}"), { kubernetes: Object.assign({}, deploymentData, { deploymentSpec: JSON.stringify(deploymentSpec), serviceSpec: JSON.stringify(serviceSpec) }) })) });
        }));
    });
}
/**
 * Default deployment data callback that reads the Dockerfile and uses EXPOSE instructions to setup
 * ingress rules.
 */
function defaultDeploymentData(p, goal, ctx, k8Deploy) {
    return __awaiter(this, void 0, void 0, function* () {
        const configuration = k8Deploy.sdm.configuration;
        const details = k8Deploy.details;
        const ns = details && details.environment ? details.environment : kubernetes_1.defaultNamespace;
        let ingress = {};
        if (yield p.hasFile("Dockerfile")) {
            const df = yield p.getFile("Dockerfile");
            const parser = require("docker-file-parser");
            const options = { includeComments: false };
            const commands = parser.parse(yield df.getContent(), options);
            const exposeCommands = commands.filter((c) => c.name === "EXPOSE");
            if (exposeCommands.length > 1) {
                throw new Error(`Unable to determine port for default ingress. Dockerfile in project ` +
                    `'${goal.repo.owner}/${goal.repo.name}' has more then one EXPOSE instruction: ` +
                    exposeCommands.map((c) => c.args).join(", "));
            }
            else if (exposeCommands.length === 1) {
                let host = "sdm.info";
                let path = `/${ns}/${goal.repo.owner}/${goal.repo.name}`;
                if (_.get(configuration, "sdm.k8.ingress.host")) {
                    host = _.get(configuration, "sdm.k8.ingress.host");
                }
                else if (_.get(configuration, "sdm.k8.context") === "minikube") {
                    // Attempt to load the minikube ip and use that to construct a hostname
                    const log = new sdm_1.StringCapturingProgressLog();
                    const result = yield automation_client_1.spawnAndWatch({
                        command: "minikube",
                        args: ["ip"],
                    }, {}, log, {
                        errorFinder: automation_client_1.SuccessIsReturn0ErrorFinder,
                        logCommand: false,
                    });
                    if (result.code === 0) {
                        host = `${goal.repo.name}.${goal.repo.owner}.${ns}.${log.log.trim()}.nip.io`;
                        path = "/";
                    }
                }
                ingress = {
                    host,
                    port: +exposeCommands[0].args[0],
                    protocol: "http",
                    path,
                };
            }
        }
        return Object.assign({ name: goal.repo.name, environment: configuration.environment, ns }, ingress);
    });
}
exports.defaultDeploymentData = defaultDeploymentData;
function getEnvironmentLabel(details) {
    if (details && details.environment && typeof details.environment === "string") {
        switch (details.environment) {
            case "testing":
                return " to `testing`";
            case "production":
                return " to `production`";
            default:
                return ` to \`${details.environment}\``;
        }
    }
    else if (details && details.environment) {
        switch (details.environment) {
            case sdm_1.StagingEnvironment:
                return " to `testing`";
            case sdm_1.ProductionEnvironment:
                return " to `production`";
            default:
                return "";
        }
    }
    else {
        return "";
    }
}
function getEnvironment(details) {
    if (details && details.environment && typeof details.environment === "string") {
        switch (details.environment) {
            case "testing":
                return sdm_1.StagingEnvironment;
            case "production":
                return sdm_1.ProductionEnvironment;
            default:
                return sdm_1.IndependentOfEnvironment;
        }
    }
    else if (details && details.environment) {
        return details.environment;
    }
    else {
        return sdm_1.IndependentOfEnvironment;
    }
}
//# sourceMappingURL=KubernetesDeploy.js.map