import * as k8 from "kubernetes-client";
import { Deployment, Ingress, Namespace, Service } from "../typings/kubernetes";
/**
 * Get Kubernetes configuration either from the creds directory or the
 * in-cluster client.
 */
export declare function getKubeConfig(context?: string): k8.ClusterConfiguration | k8.ClientConfiguration;
export declare function loadKubeConfig(): any;
/**
 * Kubernetes configuration to use to create API clients.
 */
export interface KubeConfig {
    /** Kubernetes cluster or client configuration */
    config: k8.ClusterConfiguration | k8.ClientConfiguration;
}
/**
 * Kubernetes API clients used to create/update/delete an application.
 */
export interface KubeClients {
    /** Kubernetes Core client */
    core: k8.ApiGroup;
    /** Kubernetes Extension client */
    ext: k8.ApiGroup;
    /** Kubernetes Apps client */
    apps: k8.ApiGroup;
}
/**
 * Information needed to construct resources required for creating or
 * updating an application in a Kubernetes cluster.
 */
export interface KubeApplication {
    /** Atomist workspace ID */
    workspaceId: string;
    /** Arbitrary name of environment */
    environment: string;
    /** Name of resources to create */
    name: string;
    /** Namespace to create resources in */
    ns: string;
    /** Full image name and tag for deployment pod template container */
    image: string;
    /**
     * Name of image pull secret for container image, if not provided
     * no image pull secret is provided in the pod spec.
     */
    imagePullSecret?: string;
    /**
     * Port the service listens on, if not provided no service
     * resource is created.
     */
    port?: number;
    /**
     * Ingress rule URL path, if not provided no ingress rule is
     * added.
     */
    path?: string;
    /**
     * Ingress rule hostname, if not provided none is used in the
     * ingress rule, meaning it will apply to the wildcard host, and
     * "localhost" is used when constructing the service endpoint URL.
     */
    host?: string;
    /**
     * Ingress protocol, "http" or "https".  If tslSecret is provided,
     * the default is "https", otherwise "http".
     */
    protocol?: "http" | "https";
    /** Name of TLS secret for host */
    tlsSecret?: string;
    /**
     * Stringified patch of a deployment spec for this application
     * that is parsed and overlaid on top of the default deployment
     * spec template.
     */
    deploymentSpec?: string;
    /**
     * Stringified patch of a service spec for this application that
     * is parsed and overlaid on top of the default service spec
     * template.
     */
    serviceSpec?: string;
    /**
     * Number of replicas in deployment.  May be overridden by
     * deploymentSpec.
     */
    replicas?: number;
}
/**
 * Information needed to delete resources related to an application in
 * a Kubernetes cluster.
 */
export declare type KubeDelete = Pick<KubeApplication, "name" | "ns" | "path" | "host">;
/**
 * Information needed to create an application in a Kubernetes
 * cluster.
 */
export declare type KubeApplicationRequest = KubeConfig & KubeApplication;
/**
 * Information needed to delete an application from a Kubernetes
 * cluster.
 */
export declare type KubeDeleteRequest = KubeConfig & KubeDelete;
/**
 * Create or update all the resources for an application in a
 * Kubernetes cluster if it does not exist.  If it does exist, update
 * the image in the deployment resource.
 *
 * @param req application creation request
 */
export declare function upsertApplication(upReq: KubeApplicationRequest): Promise<void>;
/**
 * Delete an application from a kubernetes cluster.  If any resource
 * requested to be deleted does not exist, it is logged but no error
 * is returned.
 *
 * @param req delete application request object
 */
export declare function deleteApplication(delReq: KubeDeleteRequest): Promise<void>;
/**
 * Create namespace resource.
 *
 * @param req Kubernetes application
 * @return kubernetes namespace resource
 */
export declare function namespaceTemplate(req: KubeApplication): Namespace;
/**
 * Create deployment patch for a repo and image.  If the request has a
 * `deploymentSpec`, it is merged into the patch created by this
 * function.
 *
 * @param req deployment template request
 * @return deployment resource patch
 */
export declare function deploymentPatch(req: KubeApplication): Partial<Deployment>;
/**
 * Create deployment for a repo and image.  If the request has a
 * `deploymentSpec`, it is merged into the patch created by this
 * function.
 *
 * @param req deployment template request
 * @return deployment resource
 */
export declare function deploymentTemplate(req: KubeApplication): Deployment;
/**
 * Create service to front a deployment for a repo and image.
 *
 * @param req service template request
 * @return service resource
 */
export declare function serviceTemplate(req: KubeApplication): Service;
/**
 * Create the URL for a deployment using the protocol, host, and tail
 * from the req object.  If host is not provided, use "localhost".  If
 * protocol is not provided, use "https" if tlsSecret is provided,
 * otherwise "http".  A forward slash, /, is appended to the tail even
 * if it is empty.
 *
 * @param req ingress request
 * @return endpoint URL for deployment service
 */
export declare function endpointBaseUrl(req: KubeApplication): string;
/**
 * Create the ingress for a deployment namespace.
 *
 * @param req ingress request
 * @return ingress template with single rule
 */
export declare function ingressTemplate(req: KubeApplication): Ingress;
/**
 * Create a patch to add a new path to the ingress rules.  If there is
 * already a rule for the path/backend/host, it returns undefined.
 *
 * @param ing ingress resource to create patch for
 * @param req ingress request
 * @return ingress patch or undefined if no need to patch
 */
export declare function ingressPatch(ing: Ingress, req: KubeApplication): Partial<Ingress>;
/**
 * Create a patch to remove a path from the ingress rules.  If the
 * path does not exist, undefined is returned.  Note that if the last
 * rule is removed, a patch having a spec with an empty rule array
 * will be returned.
 *
 * @param ing ingress resource to create patch for
 * @param req ingress request
 * @return ingress patch that removes the path, or undefined if nothing needs done.
 */
export declare function ingressRemove(ing: Ingress, req: KubeDelete): Partial<Ingress>;
