"use strict";
/*
 * Copyright © 2018 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const stringify = require("json-stringify-safe");
const k8 = require("kubernetes-client");
const _ = require("lodash");
const os = require("os");
const path = require("path");
const promiseRetry = require("promise-retry");
const error_1 = require("./error");
// tslint:disable:max-file-line-count
function configPath() {
    return process.env.KUBECONFIG || path.join(os.homedir(), ".kube", "config");
}
/**
 * Get Kubernetes configuration either from the creds directory or the
 * in-cluster client.
 */
function getKubeConfig(context) {
    let k8Config;
    const cfgPath = configPath();
    try {
        const kubeconfig = k8.config.loadKubeconfig(cfgPath);
        k8Config = k8.config.fromKubeconfig(kubeconfig, context);
    }
    catch (e) {
        automation_client_1.logger.debug(`failed to use ${cfgPath}: ${e.message}`);
        k8Config = k8.config.getInCluster();
    }
    return k8Config;
}
exports.getKubeConfig = getKubeConfig;
function loadKubeConfig() {
    const cfgPath = configPath();
    return k8.config.loadKubeconfig(cfgPath);
}
exports.loadKubeConfig = loadKubeConfig;
/**
 * Create the KubeClients structure.
 */
function getKubeClients(config) {
    const core = new k8.Core(config);
    const ext = new k8.Extensions(config);
    const apps = new k8.Apps(config);
    return { core, ext, apps };
}
function reqFilter(k, v) {
    if (k === "config" || k === "core" || k === "ext") {
        return undefined;
    }
    return v;
}
/**
 * Create or update all the resources for an application in a
 * Kubernetes cluster if it does not exist.  If it does exist, update
 * the image in the deployment resource.
 *
 * @param req application creation request
 */
function upsertApplication(upReq) {
    const clients = getKubeClients(upReq.config);
    const req = Object.assign({}, upReq, clients);
    const reqStr = stringify(req, reqFilter);
    return upsertNamespace(req)
        .then(() => upsertService(req))
        .then(() => upsertDeployment(req))
        .then(() => upsertIngress(req))
        .catch(e => Promise.reject(error_1.preErrMsg(e, `upserting '${reqStr}' failed`)));
}
exports.upsertApplication = upsertApplication;
/**
 * Delete an application from a kubernetes cluster.  If any resource
 * requested to be deleted does not exist, it is logged but no error
 * is returned.
 *
 * @param req delete application request object
 */
function deleteApplication(delReq) {
    const clients = getKubeClients(delReq.config);
    const req = Object.assign({}, delReq, clients);
    const reqStr = stringify(req, reqFilter);
    const slug = `${req.ns}/${req.name}`;
    const errs = [];
    return deleteIngress(req)
        .catch(e => errs.push(error_1.preErrMsg(e, `failed to remove rule for ${slug} from ingress`)))
        .then(() => deleteService(req))
        .catch(e => errs.push(error_1.preErrMsg(e, `failed to delete service ${slug}`)))
        .then(() => deleteDeployment(req))
        .catch(e => errs.push(error_1.preErrMsg(e, `failed to delete deployment ${slug}`)))
        .then(() => {
        if (errs.length > 0) {
            const msg = `Failed to delete application '${reqStr}': ${errs.map(e => e.message).join("; ")}`;
            automation_client_1.logger.error(msg);
            return Promise.reject(new Error(msg));
        }
        return Promise.resolve();
    });
}
exports.deleteApplication = deleteApplication;
/**
 * Create or update a namespace.
 *
 * @param req Kuberenetes application request
 */
function upsertNamespace(req) {
    return req.core.namespaces(req.ns).get()
        .then(() => automation_client_1.logger.debug(`Namespace ${req.ns} exists`), e => {
        automation_client_1.logger.debug(`Failed to get namespace ${req.ns}, creating: ${e.message}`);
        const ns = namespaceTemplate(req);
        automation_client_1.logger.debug(`Creating namespace ${req.ns} using '${stringify(ns)}'`);
        return retryP(() => req.core.namespaces.post({ body: ns }), `create namespace ${req.ns}`);
    });
}
/**
 * Create a service if it does not exist.  If req.port is false, no
 * service is created.
 *
 * @param req Kuberenetes application request
 */
function upsertService(req) {
    const slug = `${req.ns}/${req.name}`;
    if (!req.port) {
        automation_client_1.logger.debug(`Port not provided, will not create service ${slug}`);
        return Promise.resolve();
    }
    return req.core.namespaces(req.ns).services(req.name).get()
        .then(() => automation_client_1.logger.debug(`Service ${slug} exists`), e => {
        automation_client_1.logger.debug(`Failed to get service ${slug}, creating: ${e.message}`);
        let svc;
        try {
            svc = serviceTemplate(req);
        }
        catch (e) {
            automation_client_1.logger.error(e.message);
            return Promise.reject(e);
        }
        automation_client_1.logger.debug(`Creating service ${slug} using '${stringify(svc)}'`);
        return retryP(() => req.core.namespaces(req.ns).services.post({ body: svc }), `create service ${slug}`);
    });
}
/**
 * Create or updated a deployment.
 *
 * @param req Kuberenetes application request
 */
function upsertDeployment(req) {
    const slug = `${req.ns}/${req.name}`;
    return req.ext.namespaces(req.ns).deployments(req.name).get()
        .then(dep => {
        let patch;
        try {
            patch = deploymentPatch(req);
        }
        catch (e) {
            automation_client_1.logger.error(e.message);
            return Promise.reject(e);
        }
        automation_client_1.logger.debug(`Updating deployment ${slug} using '${stringify(patch)}'`);
        return retryP(() => req.ext.namespaces(req.ns).deployments(req.name).patch({ body: patch }), `patch deployment ${slug}`);
    }, e => {
        automation_client_1.logger.debug(`Failed to get deployment ${slug}, creating: ${e.message}`);
        let dep;
        try {
            dep = deploymentTemplate(req);
        }
        catch (e) {
            automation_client_1.logger.error(e.message);
            return Promise.reject(e);
        }
        automation_client_1.logger.debug(`Creating deployment ${slug} using '${stringify(dep)}'`);
        return retryP(() => req.ext.namespaces(req.ns).deployments.post({ body: dep }), `create deployment ${slug}`);
    });
}
/**
 * Create or updated an ingress with the appropriate rule for an
 * application.
 *
 * @param req Kuberenetes resource request
 */
function upsertIngress(req) {
    const slug = `${req.ns}/${req.name}`;
    if (!req.path) {
        automation_client_1.logger.debug(`Path not provided, will not upsert ingress ${slug}`);
        return Promise.resolve();
    }
    return req.ext.namespaces(req.ns).ingresses(ingressName).get()
        .then((ing) => {
        let patch;
        try {
            patch = ingressPatch(ing, req);
        }
        catch (e) {
            automation_client_1.logger.error(e.message);
            return Promise.reject(e);
        }
        if (!patch) {
            automation_client_1.logger.debug(`Ingress ${ingressName} does not need updating for ${slug}: ${stringify(ing)}`);
            return Promise.resolve();
        }
        automation_client_1.logger.debug(`Updating ingress ${ingressName} for ${slug} using '${stringify(patch)}'`);
        return retryP(() => req.ext.namespaces(req.ns).ingresses(ingressName).patch({ body: patch }), `patch ingress ${req.ns}/${ingressName} for ${slug}`);
    }, e => {
        automation_client_1.logger.debug(`Failed to get ingress ${req.ns}/${ingressName}, creating: ${e.message}`);
        const ing = ingressTemplate(req);
        automation_client_1.logger.debug(`Creating ingress ${ingressName} for ${slug} using '${stringify(ing)}'`);
        return retryP(() => req.ext.namespaces(req.ns).ingresses.post({ body: ing }), `create ingress ${req.ns}/${ingressName} for ${slug}`);
    });
}
/**
 * Delete a service if it exists.  If the resource does not exist, do
 * nothing.
 *
 * @param req Kuberenetes delete request
 */
function deleteService(req) {
    const slug = `${req.ns}/${req.name}`;
    return req.core.namespaces(req.ns).services(req.name).get()
        .then(() => {
        return retryP(() => req.core.namespaces(req.ns).services(req.name).delete({}), `delete service ${slug}`);
    }, e => automation_client_1.logger.debug(`Service ${slug} does not exist: ${e.message}`));
}
/**
 * Delete a deployment if it exists.  If the resource does not exist,
 * do nothing.
 *
 * @param req Kuberenetes delete request
 */
function deleteDeployment(req) {
    const slug = `${req.ns}/${req.name}`;
    return req.ext.namespaces(req.ns).deployments(req.name).get()
        .then(() => {
        const body = { propagationPolicy: "Background" };
        return retryP(() => req.ext.namespaces(req.ns).deployments(req.name).delete({ body }), `delete deployment ${slug}`);
    }, e => automation_client_1.logger.debug(`Deployment ${slug} does not exist: ${e.message}`));
}
/**
 * Delete a rule from an ingress if it exists.  If the rule does not
 * exist, do nothing.  If it is the last rule to be deleted, the
 * ingress is deleted.
 *
 * @param req Kuberenetes delete request
 */
function deleteIngress(req) {
    const slug = `${req.ns}/${req.name}`;
    if (!req.path) {
        automation_client_1.logger.debug(`No path provided for ${slug}, cannot delete ingress rule`);
        return Promise.resolve();
    }
    return retryP(() => req.ext.namespaces(req.ns).ingresses(ingressName).get(), "get ingress")
        .then((ing) => {
        let patch;
        try {
            patch = ingressRemove(ing, req);
        }
        catch (e) {
            automation_client_1.logger.error(e.message);
            return Promise.reject(e);
        }
        if (patch === undefined) {
            automation_client_1.logger.debug(`No changes to ingress necessary for ${slug}`);
            return Promise.resolve();
        }
        else if (patch.spec.rules.length < 1) {
            automation_client_1.logger.debug(`Last rule removed from ingress ${req.ns}/${ingressName}, deleting ingress`);
            return retryP(() => req.ext.namespaces(req.ns).ingresses(ingressName).delete({}), "delete ingress");
        }
        return retryP(() => req.ext.namespaces(req.ns).ingresses(ingressName).patch({ body: patch }), "remove path from ingress");
    }, e => automation_client_1.logger.debug(`Ingress ${req.ns}/${ingressName} does not exist: ${e.message}`));
}
const creator = `atomist.k8-automation`;
/**
 * Create namespace resource.
 *
 * @param req Kubernetes application
 * @return kubernetes namespace resource
 */
function namespaceTemplate(req) {
    const ns = {
        apiVersion: "v1",
        kind: "Namespace",
        metadata: {
            name: req.ns,
        },
    };
    return ns;
}
exports.namespaceTemplate = namespaceTemplate;
function labels(req) {
    return {
        app: req.name,
        workspaceId: req.workspaceId,
        env: req.environment,
        creator,
    };
}
/**
 * Create deployment patch for a repo and image.  If the request has a
 * `deploymentSpec`, it is merged into the patch created by this
 * function.
 *
 * @param req deployment template request
 * @return deployment resource patch
 */
function deploymentPatch(req) {
    const patch = {
        spec: {
            template: {
                spec: {
                    containers: [
                        {
                            name: req.name,
                            image: req.image,
                        },
                    ],
                },
            },
        },
    };
    if (req.replicas) {
        patch.spec.replicas = req.replicas;
    }
    if (req.deploymentSpec) {
        try {
            const depSpec = JSON.parse(req.deploymentSpec);
            _.merge(patch, depSpec);
        }
        catch (e) {
            throw new Error(`Failed to parse provided deployment spec as JSON: ${e.message}`);
        }
    }
    return patch;
}
exports.deploymentPatch = deploymentPatch;
/**
 * Create deployment for a repo and image.  If the request has a
 * `deploymentSpec`, it is merged into the patch created by this
 * function.
 *
 * @param req deployment template request
 * @return deployment resource
 */
function deploymentTemplate(req) {
    const k8ventAnnot = stringify({
        environment: req.environment,
        webhooks: [`${automation_client_1.webhookBaseUrl()}/atomist/kube/teams/${req.workspaceId}`],
    });
    const d = {
        apiVersion: "extensions/v1beta1",
        kind: "Deployment",
        metadata: {
            name: req.name,
            labels: labels(req),
        },
        spec: {
            replicas: (req.replicas || req.replicas === 0) ? req.replicas : 1,
            revisionHistoryLimit: 3,
            selector: {
                matchLabels: {
                    app: req.name,
                    workspaceId: req.workspaceId,
                },
            },
            template: {
                metadata: {
                    name: req.name,
                    labels: labels(req),
                    annotations: {
                        "atomist.com/k8vent": k8ventAnnot,
                    },
                },
                spec: {
                    containers: [
                        {
                            name: req.name,
                            image: req.image,
                            imagePullPolicy: "IfNotPresent",
                            resources: {
                                limits: {
                                    cpu: "1000m",
                                    memory: "384Mi",
                                },
                                requests: {
                                    cpu: "100m",
                                    memory: "320Mi",
                                },
                            },
                        },
                    ],
                    dnsPolicy: "ClusterFirst",
                    restartPolicy: "Always",
                },
            },
            strategy: {
                type: "RollingUpdate",
                rollingUpdate: {
                    maxUnavailable: 0,
                    maxSurge: 1,
                },
            },
        },
    };
    if (req.port) {
        d.spec.template.spec.containers[0].ports = [
            {
                name: "http",
                containerPort: req.port,
                protocol: "TCP",
            },
        ];
        const probe = {
            httpGet: {
                path: "/",
                port: "http",
                scheme: "HTTP",
            },
            initialDelaySeconds: 30,
            timeoutSeconds: 3,
            periodSeconds: 10,
            successThreshold: 1,
            failureThreshold: 3,
        };
        d.spec.template.spec.containers[0].readinessProbe = probe;
        d.spec.template.spec.containers[0].livenessProbe = probe;
    }
    if (req.imagePullSecret) {
        d.spec.template.spec.imagePullSecrets = [{ name: req.imagePullSecret }];
    }
    if (req.deploymentSpec) {
        try {
            const depSpec = JSON.parse(req.deploymentSpec);
            _.merge(d, depSpec);
        }
        catch (e) {
            throw new Error(`Failed to parse provided deployment spec as JSON: ${e.message}`);
        }
    }
    return d;
}
exports.deploymentTemplate = deploymentTemplate;
/**
 * Create service to front a deployment for a repo and image.
 *
 * @param req service template request
 * @return service resource
 */
function serviceTemplate(req) {
    const s = {
        kind: "Service",
        apiVersion: "v1",
        metadata: {
            name: req.name,
            labels: labels(req),
        },
        spec: {
            ports: [
                {
                    name: "http",
                    protocol: "TCP",
                    port: req.port,
                    targetPort: "http",
                },
            ],
            selector: {
                app: req.name,
                workspaceId: req.workspaceId,
            },
            sessionAffinity: "None",
            type: "NodePort",
        },
    };
    if (req.serviceSpec) {
        try {
            const svcSpec = JSON.parse(req.serviceSpec);
            _.merge(s, svcSpec);
        }
        catch (e) {
            throw new Error(`Failed to parse provided service spec as JSON: ${e.message}`);
        }
    }
    return s;
}
exports.serviceTemplate = serviceTemplate;
const ingressName = "atm-ingress";
/**
 * Create the URL for a deployment using the protocol, host, and tail
 * from the req object.  If host is not provided, use "localhost".  If
 * protocol is not provided, use "https" if tlsSecret is provided,
 * otherwise "http".  A forward slash, /, is appended to the tail even
 * if it is empty.
 *
 * @param req ingress request
 * @return endpoint URL for deployment service
 */
function endpointBaseUrl(req) {
    const defaultProtocol = (req.tlsSecret) ? "https" : "http";
    const protocol = (req.protocol) ? req.protocol : defaultProtocol;
    const host = (req.host) ? req.host : "localhost";
    const tail = (req.path && req.path !== "/") ? `${req.path}/` : "/";
    return `${protocol}://${host}${tail}`;
}
exports.endpointBaseUrl = endpointBaseUrl;
/**
 * Create a ingress HTTP path.
 *
 * @param req ingress request
 * @return ingress HTTP path
 */
function httpIngressPath(req) {
    const httpPath = {
        path: req.path,
        backend: {
            serviceName: req.name,
            servicePort: "http",
        },
    };
    return httpPath;
}
/**
 * Create the ingress for a deployment namespace.
 *
 * @param req ingress request
 * @return ingress template with single rule
 */
function ingressTemplate(req) {
    const httpPath = httpIngressPath(req);
    const rule = {
        http: {
            paths: [httpPath],
        },
    };
    if (req.host) {
        rule.host = req.host;
    }
    const i = {
        kind: "Ingress",
        apiVersion: "extensions/v1beta1",
        metadata: {
            name: ingressName,
            annotations: {
                "kubernetes.io/ingress.class": "nginx",
                "nginx.ingress.kubernetes.io/rewrite-target": "/",
            },
            labels: {
                ingress: "nginx",
                workspaceId: req.workspaceId,
                env: req.environment,
                creator,
            },
        },
        spec: {
            rules: [rule],
        },
    };
    if (req.tlsSecret) {
        i.spec.tls = [
            {
                secretName: req.tlsSecret,
            },
        ];
        if (req.host) {
            i.spec.tls[0].hosts = [req.host];
        }
    }
    return i;
}
exports.ingressTemplate = ingressTemplate;
/**
 * Create a patch to add a new path to the ingress rules.  If there is
 * already a rule for the path/backend/host, it returns undefined.
 *
 * @param ing ingress resource to create patch for
 * @param req ingress request
 * @return ingress patch or undefined if no need to patch
 */
// tslint:disable:cyclomatic-complexity
function ingressPatch(ing, req) {
    const httpPath = httpIngressPath(req);
    const rules = (ing && ing.spec && ing.spec.rules) ? ing.spec.rules : [];
    const ruleIndex = (req.host) ? rules.findIndex(r => r.host === req.host) : rules.findIndex(r => !r.host);
    if (ruleIndex < 0) {
        const rule = {
            http: {
                paths: [httpPath],
            },
        };
        if (req.host) {
            rule.host = req.host;
        }
        rules.push(rule);
    }
    else {
        const rule = rules[ruleIndex];
        const paths = (rule && rule.http && rule.http.paths) ? rule.http.paths : [];
        const existingPath = paths.find(p => p.path === req.path);
        if (existingPath) {
            if (existingPath.backend.serviceName !== req.name) {
                throw new Error(`Cannot use path '${req.path}' for service ${req.ns}/${req.name}, it is already ` +
                    `in use by ${existingPath.backend.serviceName}`);
            }
            automation_client_1.logger.debug(`Rule with path '${req.path}' and service ${req.ns}/${req.name} already exists`);
            return undefined;
        }
        paths.push(httpPath);
        rule.http.paths = paths;
        rules[ruleIndex] = rule;
    }
    const patch = { spec: { rules } };
    if (req.tlsSecret) {
        const tls = (ing && ing.spec && ing.spec.tls) ? ing.spec.tls : [];
        const tlsIndex = tls.findIndex(t => t.secretName === req.tlsSecret);
        if (tlsIndex < 0) {
            const t = {
                secretName: req.tlsSecret,
            };
            if (req.host) {
                t.hosts = [req.host];
            }
            tls.push(t);
        }
        else {
            if (req.host) {
                if (!tls[tlsIndex].hosts.some(h => h === req.host)) {
                    tls[tlsIndex].hosts.push(req.host);
                }
            }
        }
        patch.spec.tls = tls;
    }
    return patch;
}
exports.ingressPatch = ingressPatch;
// tslint:enable:cyclomatic-complexity
/**
 * Create a patch to remove a path from the ingress rules.  If the
 * path does not exist, undefined is returned.  Note that if the last
 * rule is removed, a patch having a spec with an empty rule array
 * will be returned.
 *
 * @param ing ingress resource to create patch for
 * @param req ingress request
 * @return ingress patch that removes the path, or undefined if nothing needs done.
 */
function ingressRemove(ing, req) {
    if (!ing || !ing.spec || !ing.spec.rules || ing.spec.rules.length < 1) {
        return undefined;
    }
    const rules = ing.spec.rules;
    const ruleIndex = (req.host) ? rules.findIndex(r => r.host === req.host) : rules.findIndex(r => !r.host);
    if (ruleIndex < 0) {
        return undefined;
    }
    const rule = rules[ruleIndex];
    const paths = rule.http.paths;
    const pathIndex = paths.findIndex(p => p.path === req.path);
    if (pathIndex < 0) {
        return undefined;
    }
    const existingPath = paths[pathIndex];
    if (existingPath.backend.serviceName !== req.name) {
        throw new Error(`Will not remove path '${req.path}' for service ${req.ns}/${req.name}, it is associated ` +
            `with service ${existingPath.backend.serviceName}`);
    }
    rules[ruleIndex].http.paths.splice(pathIndex, 1);
    if (rules[ruleIndex].http.paths.length < 1) {
        rules.splice(ruleIndex, 1);
    }
    const patch = { spec: { rules } };
    return patch;
}
exports.ingressRemove = ingressRemove;
const defaultRetryOptions = {
    retries: 5,
    factor: 2,
    minTimeout: 0.1 * 1000,
    maxTimeout: 3 * 1000,
    randomize: true,
};
/**
 * Retry Kube API call.
 */
function retryP(k, desc, options = defaultRetryOptions) {
    return promiseRetry(defaultRetryOptions, (retry, count) => {
        automation_client_1.logger.debug(`Retry ${desc} attempt ${count}`);
        return k().catch(e => {
            automation_client_1.logger.debug(`Error ${desc} attempt ${count}: ${e.message}`);
            retry(e);
        });
    })
        .catch(e => Promise.reject(error_1.preErrMsg(e, `Failed to ${desc}`)));
}
//# sourceMappingURL=api.js.map