"use strict";
/*
 * Copyright © 2018 Atomist, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const automation_client_1 = require("@atomist/automation-client");
const stringify = require("json-stringify-safe");
const kubernetes_1 = require("../typings/kubernetes");
const api_1 = require("./api");
function executeKubernetesDeploy() {
    return (goalInvocation) => __awaiter(this, void 0, void 0, function* () {
        const { configuration, context, sdmGoal, progressLog } = goalInvocation;
        const options = configuration.sdm.k8;
        const repo = sdmGoal.repo.name;
        const owner = sdmGoal.repo.owner;
        const sha = sdmGoal.sha;
        const workspaceId = context.workspaceId;
        const env = configuration.environment;
        const depName = `${workspaceId}:${env}:${owner}:${repo}:${sha}`;
        if (!sdmGoal.push.after.image) {
            const msg = `Kubernetes deploy requested for ${depName} but that commit ` +
                `has no Docker image associated with it`;
            return failDeploy(progressLog, msg);
        }
        const image = sdmGoal.push.after.image.imageName;
        progressLog.write(`Processing ${depName}`);
        let k8Config;
        try {
            k8Config = api_1.getKubeConfig(options.context);
        }
        catch (e) {
            return failDeploy(progressLog, `Failed to get kube config for ${depName}: ${e.message}`);
        }
        const kubeGoalData = validateSdmGoal(sdmGoal);
        if (!kubeGoalData) {
            return { code: 1, message: "No Kubernetes goal data found" };
        }
        const kubeApp = Object.assign({}, kubeGoalData, { workspaceId,
            image, ns: kubeGoalData.ns || kubernetes_1.defaultNamespace });
        progressLog.write(`Deploying ${depName} to Kubernetes`);
        const upsertReq = Object.assign({}, kubeApp, { config: k8Config });
        try {
            yield api_1.upsertApplication(upsertReq);
        }
        catch (e) {
            return failDeploy(progressLog, `Failed to deploy ${depName} to Kubernetes: ${e.message}`);
        }
        const message = `Successfully deployed ${depName} to Kubernetes`;
        progressLog.write(message);
        const description = `Deployed to Kubernetes namespace \`${kubeApp.ns}\``;
        const label = `Kubernetes ${env}:${kubeApp.ns}`;
        const targetUrls = (kubeApp.path && kubeApp.host) ? [{ label, url: api_1.endpointBaseUrl(kubeApp) }] : undefined;
        return { code: 0, message, description, targetUrls };
    });
}
exports.executeKubernetesDeploy = executeKubernetesDeploy;
/**
 * Validate the SDM goal has all necessary data.  It will return the
 * KubeApplication data if the deployment should proceed.  It will ll
 * return undefined if nothing should be deployed.
 *
 * @param sdmGoal SDM goal for Kubernetes application deployment
 * @return valid KubeApplication if something should be deployed,
 *         undefined if nothing should be deployed
 */
function validateSdmGoal(sdmGoal) {
    if (!sdmGoal.data) {
        automation_client_1.logger.debug(`SDM goal data property is false, cannot deploy: '${stringify(sdmGoal)}'`);
        return undefined;
    }
    let sdmData;
    try {
        sdmData = JSON.parse(sdmGoal.data);
    }
    catch (e) {
        automation_client_1.logger.debug(`Failed to parse SDM goal data '${sdmGoal.data}' as JSON: ${e.message}`);
        return undefined;
    }
    if (!sdmData.kubernetes) {
        automation_client_1.logger.debug(`SDM goal data kubernetes property is false, cannot deploy: '${stringify(sdmData)}'`);
        return undefined;
    }
    const kubeApp = sdmData.kubernetes;
    if (!kubeApp.name) {
        automation_client_1.logger.debug(`SDM goal data kubernetes name property is false, cannot deploy: '${stringify(sdmData)}'`);
        return undefined;
    }
    return kubeApp;
}
exports.validateSdmGoal = validateSdmGoal;
/**
 * Report and return failure.
 *
 * @param message informative error message
 * @return a failure handler result using the provided error message
 */
function failDeploy(log, message) {
    log.write(message);
    automation_client_1.logger.error(message);
    return { code: 1, message };
}
//# sourceMappingURL=deploy.js.map